#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "sqlite3.h"
#include "linkface_sdk_v2.h"


#define LINK_FACE_DB_NAME   "linkface.db"


static sqlite3 *g_hdr_db = NULL;

static inline int _get_number_length(int base)
{
    int ret = 0;

    do {
        ret += 1;
        base /= 10;
    } while(base > 0);

    return ret;
}

static int create_user_info_table()
{
    const char *sql = "CREATE TABLE if not exists user_info ( \
                        person_id   VARCHAR(128) PRIMARY KEY NOT NULL, \
                        name            VARCHAR(128), \
                        group_ids       VARCHAR(1024), \
                        group_id_count  INT, \
                        timestamp_ms    INT, \
                        extra_info      VARCHAR(2048))";
	int		ret = 0;

	ret = sqlite3_exec(g_hdr_db, sql, NULL, NULL, NULL);
	return ret;

}

static int create_image_info_table()
{
    const char *sql = "CREATE TABLE if not exists image_info ( \
                        id          INTEGER PRIMARY KEY autoincrement, \
                        person_id     VARCHAR(128), \
                        faceimg_md5   VARCHAR(64), \
                        faceimg_size  INT, \
                        timestamp_ms    INT)";
	int		ret = 0;

	ret = sqlite3_exec(g_hdr_db, sql, NULL, NULL, NULL);
	return ret;
}

int init_db()
{
    int ret = 0;

    ret = sqlite3_open(LINK_FACE_DB_NAME, &g_hdr_db);
    if(ret != 0) {
        printf("failed to open db.\n");
        return -1;
    }

    ret += create_image_info_table();
    ret += create_user_info_table();

    return ret;
}

int deinit_db()
{
    if(g_hdr_db != NULL) {
        sqlite3_close(g_hdr_db);
    }
    return 0;
}


static char *_generate_group_info(linkface_user_info *info)
{
    char *ret = NULL;
    int     i = 0;

    ret = calloc(1, (MAX_COUNT_GROUP_ID+1)*LENGTH_GROUP_ID + 1);
    if(ret == NULL) {
        return NULL;
    }

    for(i = 0; i < info->group_id_count; i++) {
        if(strlen(info->group_ids[i]) > 0) {
            strcat(ret, info->group_ids[i]);

            if(i != info->group_id_count - 1) {
                strcat(ret, ",");
            }
        }
    }

    return ret;
}

#define FMT_ADD_USER_INFO "INSERT INTO user_info" \
							"(person_id, name, group_ids, group_id_count, timestamp_ms, extra_info) VALUES " \
							"('%s', '%s', '%s', %d, %lld, '%s')"
int add_user_info(linkface_user_info *info)
{
	int  sql_len = 0;
	int  ret 	 = 0;
    char *group  = NULL;
	char *sql	 = NULL;

	if(info == NULL) {
		return -1;
	}

    group = _generate_group_info(info); 
    if(group == NULL) {
        return -1;
    }

	sql_len = strlen(FMT_ADD_USER_INFO) + strlen(info->person_id) + \
				strlen(info->name) + strlen(group)+ \
				_get_number_length(info->group_id_count)+ _get_number_length(info->timestamp_ms)+ \
				strlen(info->extra_info) + 1;

	sql = calloc(1, sql_len);
	if(sql == NULL) {
        free(group);
		return -1;
	}
	snprintf(sql, sql_len, FMT_ADD_USER_INFO, info->person_id,
							info->name, group,
							info->group_id_count, 
                            info->timestamp_ms, 
                            info->extra_info);

	ret = sqlite3_exec(g_hdr_db, sql, NULL, NULL, NULL);
    free(sql);
    free(group);
	return ret;
}

#define FMT_DEL_USER_INFO "DELETE FROM user_info WHERE person_id = %s"
int del_user_info(const char *person_id)
{
	char *sql	 = NULL;
	int  sql_len = 0;
	int  ret 	 = 0;	
	
    if(person_id == NULL) {
		return -1;
	}

	sql_len = strlen(FMT_DEL_USER_INFO) + strlen(person_id) + 1;

	sql = calloc(1, sql_len);
	if(sql == NULL) {
		return -1;
	}
	snprintf(sql, sql_len, FMT_DEL_USER_INFO, person_id);

	ret = sqlite3_exec(g_hdr_db, sql, NULL, NULL, NULL);
	return ret;
}

static int _cb_query_user_info(void *usr_data, int count, char **data, char **columns)
{
    int i = 0;
    linkface_user_info *info = (linkface_user_info *)usr_data;

    if(count < 5 || data == NULL || columns == NULL) {
        return 0;
    }
    memset(info, 0, sizeof(linkface_user_info));
    for(i = 0; i < count; i++) {
        if(data[i] == NULL)
            continue;

        if(strcmp(columns[i], "person_id") == 0) {
            strncpy(info->person_id, data[i], sizeof(info->person_id) - 1);
        } else if(strcmp(columns[i], "name") == 0) {
            strncpy(info->name, data[i], sizeof(info->name) - 1);
        } else if(strcmp(columns[i], "group_id_count") == 0) {
            info->group_id_count = atoi(data[i]);
        } else if(strcmp(columns[i], "timestamp_ms") == 0) {
            info->timestamp_ms = atoll(data[i]);
        } else if(strcmp(columns[i], "extra_info") == 0) {
            strncpy(info->extra_info, data[i], sizeof(data[i]) - 1);
        } 

    }
    return 0;
}

#define FMT_SQL_QUERY_USER_INFO_FILTER "SELECT * FROM user_info where person_id = '%s'"
int query_user_info(const char *person_id,  linkface_user_info *info)
{
    char *sql       = NULL;
    int  sql_len    = 0;
    int  ret        = 0;
    
    if(person_id == NULL || info == NULL) {
        return -1;
    }

	sql_len = strlen(FMT_SQL_QUERY_USER_INFO_FILTER) + strlen(person_id) + 1;
    sql     = calloc(1, sql_len);
    if(sql == NULL) {
        printf("memory malloc error.");
        return -1;
    }
    snprintf(sql, sql_len, FMT_SQL_QUERY_USER_INFO_FILTER, person_id);


    if(0 != sqlite3_exec(g_hdr_db, sql, _cb_query_user_info, info, NULL)){
        printf("failed to exec sql.\n");
        ret = -2;
    }
    free(sql);

    return ret;
}

#define FMT_SQL_POLLING_USER_INFO_FILTER "SELECT * FROM user_info where timestamp_ms <= %lld"
typedef struct {
    void    (*cb_sdk)(const linkface_user_info *info, void *user);
    void    *user;
} param_polling_user_info;

static int _cb_polling_user_info(void *usr_data, int count, char **data, char **columns)
{
    int i = 0;
    param_polling_user_info *info = (param_polling_user_info *)usr_data;
    linkface_user_info      user;

    if(count < 5 || data == NULL || columns == NULL) {
        return 0;
    }
    memset(&user, 0, sizeof(user));
    for(i = 0; i < count; i++) {
        if(data[i] == NULL)
            continue;

        if(strcmp(columns[i], "person_id") == 0) {
            strncpy(user.person_id, data[i], sizeof(user.person_id) - 1);
        } else if(strcmp(columns[i], "name") == 0) {
            strncpy(user.name, data[i], sizeof(user.name) - 1);
        } else if(strcmp(columns[i], "group_id_count") == 0) {
            user.group_id_count = atoi(data[i]);
        } else if(strcmp(columns[i], "timestamp_ms") == 0) {
            user.timestamp_ms = atoll(data[i]);
        } else if(strcmp(columns[i], "extra_info") == 0) {
            strncpy(user.extra_info, data[i], sizeof(data[i]) - 1);
        } 
    }
    info->cb_sdk(&user, info->user);

    return 0;
}

int polling_user_info(long long timestamp, void (*cb_sdk)(const linkface_user_info *info, void *user), void *user_sdk)
{
    char *sql    = NULL;
    int  sql_len = 0;
    int  ret     = 0;
    
    param_polling_user_info info;

	sql_len = strlen(FMT_SQL_POLLING_USER_INFO_FILTER) + _get_number_length(timestamp) + 1;
    sql     = calloc(1, sql_len);
    if(sql == NULL) {
        printf("memory malloc error.");
        return -1;
    }
    info.cb_sdk = cb_sdk;
    info.user = user_sdk;

    snprintf(sql, sql_len, FMT_SQL_POLLING_USER_INFO_FILTER, timestamp);
    if(0 != sqlite3_exec(g_hdr_db, sql, _cb_polling_user_info, &info, NULL)){
        printf("failed to exec sql.\n");
    }
    free(sql);

    return ret;

}

#define FMT_ADD_IMAGE_INFO "INSERT INTO image_info" \
							"(person_id, faceimg_md5, faceimg_size, timestamp_ms) VALUES " \
							"('%s', '%s', %d, %lld)"
int add_image_info(linkface_image_info *info)
{
	char *sql	 = NULL;
	int  sql_len = 0;
	int  ret 	 = 0;	
	if(info == NULL) {
		return -1;
	}

	sql_len = strlen(FMT_ADD_IMAGE_INFO) + strlen(info->person_id) + \
				strlen(info->faceimg_md5) + _get_number_length(info->faceimg_size)+ \
				_get_number_length(info->timestamp_ms) + 1;

	sql = calloc(1, sql_len);
	if(sql == NULL) {
		return -1;
	}
	snprintf(sql, sql_len, FMT_ADD_IMAGE_INFO, info->person_id,
							info->faceimg_md5, info->faceimg_size,
                            info->timestamp_ms);

	ret = sqlite3_exec(g_hdr_db, sql, NULL, NULL, NULL);
    free(sql);
	return ret;

}

#define FMT_DEL_IMAGE_INFO "DELETE FROM image_info WHERE person_id = %s"
int del_image_info(const char *person_id)
{
	char *sql	 = NULL;
	int  sql_len = 0;
	int  ret 	 = 0;	
	
    if(person_id == NULL) {
		return -1;
	}

	sql_len = strlen(FMT_DEL_IMAGE_INFO) + strlen(person_id) + 1;

	sql = calloc(1, sql_len);
	if(sql == NULL) {
		return -1;
	}
	snprintf(sql, sql_len, FMT_DEL_IMAGE_INFO, person_id);

	ret = sqlite3_exec(g_hdr_db, sql, NULL, NULL, NULL);
	return ret;
}

static int _cb_query_image_info(void *usr_data, int count, char **data, char **columns)
{
    int i = 0;
    linkface_image_info *info = (linkface_image_info *)usr_data;

    if(count < 4 || data == NULL || columns == NULL) {
        return 0;
    }
    memset(info, 0, sizeof(linkface_image_info));
    for(i = 0; i < count; i++) {
        if(data[i] == NULL)
            continue;

        if(strcmp(columns[i], "person_id") == 0) {
            strncpy(info->person_id, data[i], sizeof(info->person_id) - 1);
        } else if(strcmp(columns[i], "faceimg_md5") == 0) {
            strncpy(info->faceimg_md5, data[i], sizeof(info->faceimg_md5) - 1);
        } else if(strcmp(columns[i], "faceimg_size") == 0) {
            info->faceimg_size= atoi(data[i]);
        } else if(strcmp(columns[i], "timestamp_ms") == 0) {
            info->timestamp_ms = atoll(data[i]);
        } 
    }
    return 0;
}

#define FMT_SQL_QUERY_IMAGE_INFO_FILTER "SELECT * FROM image_info where person_id = '%s'"
int query_image_info(const char *person_id,  linkface_image_info *info)
{
    char *sql    = NULL;
    int  sql_len = 0;
    int  ret     = 0;

    if(person_id == NULL || info == NULL) {
        return -1;
    }

	sql_len = strlen(FMT_SQL_QUERY_IMAGE_INFO_FILTER) + strlen(person_id) + 1;
    sql     = calloc(1, sql_len);
    if(sql == NULL) {
        printf("memory malloc error.");
        return -2;
    }
    snprintf(sql, sql_len, FMT_SQL_QUERY_IMAGE_INFO_FILTER, person_id);


    if(0 != sqlite3_exec(g_hdr_db, sql, _cb_query_image_info, info, NULL)){
        printf("failed to exec sql.\n");
        ret = -3;
    }
    free(sql);

    return ret;
}

#define FMT_SQL_POLLING_IMAGE_INFO_FILTER "SELECT * FROM image_info where timestamp_ms <= %lld"
typedef struct {
    void    (*cb_sdk)(const linkface_image_info *info, void *user);
    void    *user;
} param_polling_image_info;

static int _cb_polling_image_info(void *usr_data, int count, char **data, char **columns)
{
    int i = 0;
    param_polling_image_info *info = (param_polling_image_info *)usr_data;
    linkface_image_info      user;

    if(count < 5 || data == NULL || columns == NULL) {
        return 0;
    }
    memset(&user, 0, sizeof(user));
    for(i = 0; i < count; i++) {
        if(data[i] == NULL)
            continue;

        if(strcmp(columns[i], "person_id") == 0) {
            strncpy(user.person_id, data[i], sizeof(user.person_id) - 1);
        } else if(strcmp(columns[i], "faceimg_md5") == 0) {
            strncpy(user.faceimg_md5, data[i], sizeof(user.faceimg_md5) - 1);
        } else if(strcmp(columns[i], "faceimg_size") == 0) {
            user.faceimg_size = atoi(data[i]);
        } else if(strcmp(columns[i], "timestamp_ms") == 0) {
            user.timestamp_ms = atoll(data[i]);
        }
    }
    info->cb_sdk(&user, info->user);

    return 0;
}

int polling_image_info(long long timestamp, void (*cb_sdk)(const linkface_image_info *info, void *user), void *user_sdk)
{
    char *sql       = NULL;
    int  sql_len    = 0;
    int  ret        = 0;

    param_polling_image_info info;

	sql_len = strlen(FMT_SQL_POLLING_IMAGE_INFO_FILTER) + _get_number_length(timestamp) + 1;
    sql     = calloc(1, sql_len);
    if(sql == NULL) {
        printf("memory malloc error.");
        return -1;
    }
    info.cb_sdk = cb_sdk;
    info.user = user_sdk;

    snprintf(sql, sql_len, FMT_SQL_POLLING_IMAGE_INFO_FILTER, timestamp);
    if(0 != sqlite3_exec(g_hdr_db, sql, _cb_polling_image_info, &info, NULL)){
        printf("failed to exec sql.\n");
        ret = -3;
    }
    free(sql);

    return ret;

}

static int _cb_query_tab_count(void *usr_data, int count, char **data, char **columns)
{
	int *cnt = (int *)usr_data;

	*cnt = atoi(data[0]);
	
	return 0;
}

#define FMT_SQL_QUERY_IMAGE_COUNT "SELECT count(*) from image_info " 

int query_image_count()
{
    char *sql 	= FMT_SQL_QUERY_IMAGE_COUNT;
	int	count	= -1;

	sqlite3_exec(g_hdr_db, sql, _cb_query_tab_count, &count, NULL);

	return count;
}

int clear_data()
{
    char *sql_del_image = "DELETE from image_info";
    char *sql_del_user  = "DELETE from user_info";
    int     ret         = 0;

	ret = sqlite3_exec(g_hdr_db, sql_del_image, NULL, NULL, NULL);
	ret += sqlite3_exec(g_hdr_db, sql_del_user, NULL, NULL, NULL);

    return ret;
}


